
## COCO endpoints ##

# return a list of all inventories
def get_inventories
  Log.info "getting all inventories"
  coco.get "/v0/cc/admin/inventories"
end

# return a list of all inventories
def get_inventories_for_partner partner_id
  Log.info "getting all inventories for partner_id: #{partner_id}"
  coco.get "/v0/cc/admin/#{partner_id}/inventories"
end

# get an inventory and all modules in it
def get_inventory inventory_id
  Log.info "Getting inventory - inventory_id: #{inventory_id}"
  coco.get "/v0/cc/inventory/#{inventory_id}"
end

# get an inventory and all modules in it
def get_inventory_admin partner_id, inventory_id, query_params={}
  Log.info "Admin getting inventory - partner_id: #{partner_id} - inventory_id: #{inventory_id}"
  coco.get "/v0/cc/admin/#{partner_id}/#{inventory_id}?#{URI.encode_www_form query_params}"
end

# get a inventory module
def get_inventory_module inventory_id, module_id
  Log.info "Getting module - inventory_id: #{inventory_id} - module_id: #{module_id}"
  coco.get "/v0/cc/inventory/#{URI.encode inventory_id}/module/#{URI.encode module_id}"
end

# get a single inventory module
def get_inventory_module_admin partner_id, inventory_id, module_id
  Log.info "Admin getting module - partner_id: #{partner_id} - inventory_id: #{inventory_id} - module_id: #{module_id}"
  coco.get "/v0/cc/admin/#{partner_id}/#{URI.encode inventory_id}/module/#{URI.encode module_id}"
end

# get the transaction log of an inventory,
def get_inventory_log partner_id, inventory_id, params={}
  params = {
    'start' => days_from_now(-7),
    'coalesce' => false,
  }.merge(params)
  # format the timestamp if we need to
  if params['start'].respond_to? :utc
    params['start'] = params['start'].utc.iso8601(3)
  end
  Log.info "Getting inventory log '#{inventory_id}:#{partner_id}'"
  coco.post "/v0/cc/admin/#{partner_id}/#{inventory_id}/log", params
end

# Make a new inventory
def create_inventory_admin partner_id, new_inventory_id=nil
  new_inventory_id = "KGriftTestInventory_#{random_string(16)}" if new_inventory_id.nil?
  Log.info "Admin creating an inventory - partner_id: #{partner_id}"
  # response = create_modules_admin partner_id, new_inventory_id
  response = coco.post "/v0/cc/admin/#{partner_id}/#{URI.encode new_inventory_id}/inventory", ''

  {
    'id' => new_inventory_id,
    'inventory_id' => new_inventory_id,
    'partner_id' => partner_id,
    'inventory' => response,
  }
end

def ensure_inventory_exists partner_id, inventory_id
  create_inventory_admin partner_id, inventory_id
rescue Grifter::RequestException => e
  # CoCo returns a 422 if the inventory already exists
  # any other response code is not allowed, but we'll catch 422
  raise e unless e.code.to_i == 422
end

# delete an inventory
def delete_inventory_admin partner_id, inventory_id
  Log.info "Admin deleting an inventory - partner_id: #{partner_id}"
  coco.delete "/v0/cc/admin/#{partner_id}/#{inventory_id}/inventory"
end

def ensure_inventory_does_not_exist partner_id, inventory_id
  delete_inventory_admin partner_id, inventory_id
rescue Grifter::RequestException => e
  # if the inventory does not an exist a 404 is returned.  For our purposes, that is fine
  # so we'll allow the request to fail if 404 comes back
  raise e unless e.code.to_i == 404
end

def delete_all_inventories_for_partner partner_id
  invs = get_inventories_for_partner partner_id
  inv_ids = invs['partner_inventories'].map{|i| i['inventory_id']}
  inv_ids.each do |inv_id|
    delete_inventory_admin partner_id, inv_id
  end
end

# build out a basic module object, ready to be created directly
# or you can merge in customizations
def test_module_object params={}
  params = {
    "module_id" => 'KGrift test module ' + random_uuid,
    "module_name" => random_string,
    "module_version" => random_string,
    "module_type" => ['instruction', 'assessment', 'both'].sample,
    "module_url" => "http://kgriftlms.com/#{random_string}",
    'taxons' => [],
    'module_metadata' => [],
    'containers' => [],
    'notification_url' => nil,
    'content_summary' => "TESTING: #{random_string}",
  }.merge(params)
end

# create a batch of modules against some inventory
def create_modules inventory_id, modules=[]
  modules << test_module_object if modules.empty?
  Log.info "Creating '#{modules.length}' modules for inventory_id: #{inventory_id}"
  coco.post "/v0/cc/inventory/#{inventory_id}/modules", modules
end

# create a batch of modules against some inventory
def create_modules_admin partner_id, inventory_id, modules=[]
  modules << test_module_object if modules.empty?
  Log.info "Admin creating '#{modules.length}' modules for inventory - partner_id: #{partner_id} - inventory_id: #{inventory_id}"
  coco.post "/v0/cc/admin/#{partner_id}/#{inventory_id}/modules", modules
end

# delete a module
def delete_module inventory_id, module_id
  Log.info "Deleting inventory module - inventory_id: #{inventory_id} - module_id: #{module_id}"
  #grifter doesn't actually let you put a body on a delete if you use coco.delete
  #but we can get around that like so:
  # coco.do_request :delete, "/v0/cc/inventory/#{inventory_id}/module", { 'module_id' => module_id }
  coco.delete "/v0/cc/inventory/#{inventory_id}/module/#{URI.encode module_id}"
end

# delete a module as a knerd
def delete_module_admin partner_id, inventory_id, module_id
  Log.info "Admin deleting inventory module - partner_id: #{partner_id} - inventory_id: #{inventory_id} - module_id: #{module_id}"
  #grifter doesn't actually let you put a body on a delete if you use coco.delete
  #but we can get around that like so:
  # coco.do_request :delete, "/v0/cc/admin/#{partner_id}/#{inventory_id}/module", { 'module_id' => module_id }
  # coco.delete "/v0/cc/admin/#{partner_id}/#{inventory_id}/module/#{URI.encode module_id}"
  coco.do_request :delete, "/v0/cc/admin/#{partner_id}/#{inventory_id}/module/#{URI.encode module_id}", {}
end

def update_modules inventory_id, modules
  Log.info "Updating '#{modules.length}' modules for inventory - inventory_id: #{inventory_id}"
  coco.put "/v0/cc/inventory/#{inventory_id}/modules", modules
end

def update_module inventory_id, mod
  update_modules inventory_id, [mod]
end

def update_modules_admin partner_id, inventory_id, modules
  Log.info "Updating '#{modules.length}' modules for inventory - partner_id: #{partner_id} - inventory_id: #{inventory_id}"
  coco.put "/v0/cc/admin/#{partner_id}/#{inventory_id}/modules", modules
end

def mark_module_mapped partner_id, inventory_id, module_id, module_version
  Log.info "Marking module mapped for inventory - partner_id: #{partner_id} - inventory_id: #{inventory_id} - module_id: #{module_id} - module_version: #{module_version}"
  coco.put "/v0/cc/admin/#{partner_id}/#{inventory_id}/map", {
    'module_id' => module_id,
    'module_version' => module_version,
  }
end



# CoCo stats endpoints
#

def get_partner_stats partner_id
  Log.info "getting coco partner stats - partner_id: #{partner_id}"
  coco.get "/v0/cc/stats/#{partner_id}"
end

def get_partner_stats_sum partner_id
  Log.info "getting coco partner stats sum - partner_id: #{partner_id}"
  coco.get "/v0/cc/stats/#{partner_id}/sum"
end

def get_inventory_stats partner_id, inventory_id
  Log.info "getting inventory stats - partner_id: #{partner_id} - inventory_id: #{inventory_id}"
  # coco.post "/v0/cc/stats/#{partner_id}/#{inventory_id}", {}
  coco.get "/v0/cc/stats/#{partner_id}/#{inventory_id}"
end


def get_inventory_last_updated partner_id, inventory_id
  Log.info "Admin getting inventory last updated - partner_id: #{partner_id} - inventory_id: #{inventory_id}"
  stats = get_inventory_stats partner_id, inventory_id
  stats['last_updated']
end

def get_inventory_total_modules partner_id, inventory_id
  Log.info "Admin getting inventory total modules - partner_id: #{partner_id} - inventory_id: #{inventory_id}"
  stats = get_inventory_stats partner_id, inventory_id
  stats['num_modules']
end

