require 'set'

def get_goals learning_instance_id, params={}
  Log.info "Getting all goals for learning instance: '#{learning_instance_id}' , with params '#{params.keys.join(', ')}'"
  kapi.get "/v0/learning-instances/#{learning_instance_id}/goals?" + URI.encode_www_form(params)
end

def get_goal learning_instance_id, goal_id
  Log.info "Getting goal '#{goal_id}' for learning instance '#{learning_instance_id}'"
  kapi.get "/v0/learning-instances/#{learning_instance_id}/goals/#{goal_id}"
end
alias :get_single_goal :get_goal

def get_realized_goal learning_instance_id, goal_id
  Log.info "Getting realized goal '#{goal_id}' for learning instance '#{learning_instance_id}'"
  kapi.get "/v0/learning-instances/#{learning_instance_id}/realized-goals/#{goal_id}"
end

def validate_goal goal_id, learning_instance_id
  Log.info "Validating a goal with goal id #{goal_id} on learning_instance #{learning_instance_id}"
  kapi.get "/v0/learning-instances/#{learning_instance_id}/goals/#{goal_id}/validate"
end


# Updating a goal does a full over-writing PUT.  Anything not included from the original goal
# will be removed
def update_goal learning_instance_id, goal_id, overrides={}, params={}
  Log.info "Updating a goal for learning instance: '#{learning_instance_id}' and goal: '#{goal_id}' , with params '#{params.keys.join(', ')}'"
  kapi.put "/v0/learning-instances/#{learning_instance_id}/goals/#{goal_id}?" + URI.encode_www_form(params), overrides
end

#create goal will wait twice.
#First it will wait for the goal post to succeed
#We need to do this, because a learning instance must write to kraphs
#before we can make a goal off it, and that can take up to 60 seconds
#Second it will wait for the goal to write.  Again kraphs asynchronously
#writes goals, so we have to wait for that to happen before our tests can depend on it
def create_goal learning_instance_id, overrides={}, params={}
  Log.info "Creating a goal for learning instance: '#{learning_instance_id}' , with params '#{params.keys.join(', ')}'"
  kapi.post "/v0/learning-instances/#{learning_instance_id}/goals?" + URI.encode_www_form(params), {
    'target_modules' => [],
      'recommendable_modules' => [],
      'max_recommendation_size' => 3,
      'start_date' => '2013-04-12T17:00:00.000Z'
  }.merge(overrides)
end

def create_goal_non_waiting learning_instance_id, overrides={}, params={}
  Log.info "Creating a goal for learning instance: '#{learning_instance_id}' , with params '#{params.keys.join(', ')}'"
  kapi.post "/v0/learning-instances/#{learning_instance_id}/goals?" + URI.encode_www_form(params), {
    'target_modules' => [],
      'recommendable_modules' => [],
      'max_recommendation_size' => 3,
      'start_date' => '2013-04-12T17:00:00.000Z'
  }.merge(overrides)
end


#this convenience method builds out the module lists from a graph
#which avoids building complicated payloads in tests that are not
# explicitly concerned with the complicated bits of the payload
def create_goal_from_graph graph, learning_instance_id=nil, overrides={}, params={}
  Log.info "Making a goal with random modules picked from a graph"
  nodes = graph['nodes'].clone
  Log.info nodes

  module_ids = Set.new find_all_assessment_atoms(graph)
  modules = nodes.select{|n| module_ids.member?n['id']}

  Log.info modules
  payload_overrides = {
    'target_modules' => [
      { "module_id" => modules.shift['id'],
        "target_score" => 1.0,
        "target_date" => days_from_now(1)
    },
    { "module_id" => modules.shift['id'],
      "target_score" => 1.0,
      "target_date" => days_from_now(2)
    },
    ],
    'recommendable_modules' => modules.map{|m| { "module_id" => m['id'] } }[0,5],
    'max_recommendation_size' => 2,
  }.merge(overrides)

  unless learning_instance_id
    learning_instance_id = create_learning_instance 'graph_id' => graph['id']
  end

  create_goal learning_instance_id, payload_overrides, params
end

def create_goals_from_graph_yaml yaml_file_path
  #Load yaml file
  file_data = YAML.load_file(yaml_file_path)

  #Cache graph locally
  graph = get_graph_and_cache_it file_data['graph_id']

  #Loop over all goals
  output = Hash.new()
  goals = file_data['goals']
  goals.each do |goal_name, goal_def|
    Log.info "Processing goal: #{goal_name}"

    #Initialize output variable
    output[goal_name] = Hash.new()

    #Extract input parameters
    target_module_ids = goal_def.fetch('target_module_ids',[])
    recommendable_taxon_ids = goal_def.fetch('recommendable_taxon_ids',[])
    recommendable_module_ids = goal_def.fetch('recommendable_module_ids',[])

    #Define target module
    if target_module_ids.empty?
      target_module_ids = []
      target_concept = find_prerequisite_concept graph
      target_module_ids << find_associated_assessing_bndl(graph,target_concept)
      target_module_ids << find_associated_assessing_atom(graph,target_concept)
    end
    Log.info "Target modules: #{target_module_ids}"

    #Construct target modules hash
    target_modules = []
    target_modules = target_module_ids.map do |mod|
      {
        'module_id' => mod,
        'target_date' => days_from_now(30),
        'target_score' => 0.9,
      }
    end

    #Define recommendable taxons and modules
    Log.info "Setting recommendable modules"
    rec_modules = recommendable_taxon_ids.map{ |taxon_id| {'taxon_id' => taxon_id} }
    rec_modules += recommendable_module_ids.map{ |module_id| {'module_id' => module_id} }
    if rec_modules.empty?
      Log.info "No recommendable modules or taxons define; setting all modules as recommendable"
      all_modules = find_all_recommendable_module_ids graph
      rec_modules = all_modules.map{|m| {'module_id' => m}}
    end

    #Define goal object
    goal_obj = {
      'target_modules' => target_modules,
      'max_recommendation_size' => (rec_modules.length < 3 ? rec_modules.length : 3),
      'start_date' => datetime,
      'recommendable_modules' => rec_modules,
    }
    Log.info "Creating goal with parameters #{goal_obj}"

    #Create learning instance and goal
    output[goal_name]['learning_instance_id'] = (create_learning_instance 'graph_id' => graph['id'])['id']
    output[goal_name]['goal_id'] = (create_goal output[goal_name]['learning_instance_id'], goal_obj)['id']
  end

  #Output learning instance IDs and goal IDs to screen/file
  output_file = File.join(File.dirname(yaml_file_path), File.basename(yaml_file_path, '.yml')+'_output'+'.yml')
  File.open(output_file, 'w') {|f| f.write(output.to_yaml) }
end

# GROUP GOALS

def create_group_goal learning_instance_id, params
  Log.debug "Creating a group goal for learning instance: #{learning_instance_id}"
  kapi.post "/v0/learning-instances/#{learning_instance_id}/group-goals", params
end

def update_group_goal learning_instance_id, goal_id, params
  Log.debug "Updating a group goal for learning instance: #{learning_instance_id} goal #{goal_id}"
  kapi.put "/v0/learning-instances/#{learning_instance_id}/group-goals/#{goal_id}", params
end

def create_random_valid_group_goal_object graph, options={}

  options = {
    target_type: :lref,
    num_targets: 1,
  }.merge(options)

  # find all all the recommendable modules that have a name
  potential_target_values = if options[:target_type].to_sym == :mref
                              modules = find_all_assessment_modules_full(graph)
                              # pull the mref eids out of the nodes
                              modules.map{|m| m['temp_id'] =~ /^.ref-/ ? m['temp_id'] : "mref-#{m['temp_id']}"  }
                            else
                              los = find_all_learning_objectives(graph)
                              los.map{|lo| lo['temp_id'] =~ /^.ref-/ ? lo['temp_id'] : "lref-#{lo['temp_id']}" }
                            end

  # pick a random set of target modules/learning objectives
  targets = potential_target_values.shuffle[0...options[:num_targets]]

  # make the goal object
  goal_obj = {
    "name" => "TestGroupGoal #{random_string}",
    "targets" =>  targets,
    "scope" => {},
  }

  goal_obj
end

