"""
Grifts to setup state to test recommendations.

Use to setup tests verifying rec contracts. The terminology used is specified in the beginning of
the Recommendations Contracts document:
https://docs.google.com/document/d/1s0gBAPbBKdCUmkBmLh36tQk1-s9B7VD14sz22vWm4Ns/
"""

MAX_RANDOM_HISTORY_EVENTS=20

def setup_test_scenario scenario, graph_type
    set_account(:partner)
    test_data = {}
    #get graph, create learning instance/goal/student/registration
    test_data['graph'] = (get_test_data :simple_graphs)[graph_type]
    test_data['learning_instance'] = create_learning_instance graph_id: test_data['graph']['id']
    test_data['goal'] = generate_specified_goal test_data['graph'], test_data['learning_instance']['id'], scenario['goal'], scenario['target_date_offset']
    test_data['student'] = add_new_student_to_learning_instance test_data['learning_instance']['id']

    generate_student_history scenario['history'], test_data

    test_data
end

#Generate student history of a specified type-- currently either None or Any (random)
def generate_student_history history_type, test_data
    r = Random.new
    if ENV.has_key? "REC_SEED"
        r = Random.new ENV['REC_SEED']
    end

    case history_type
    when "None"
        return
    when "Any"
        num_events = r.rand(MAX_RANDOM_HISTORY_EVENTS)
        num_events.times do
            send_random_event_on_graph test_data['student']['registration_id'], test_data['graph']
        end
    end
end

#Default goal settings
DEFAULT_TARGET_SCORE = 0.7
DEFAULT_TARGET_DATE_OFFSET = 7
def generate_specified_goal graph, learning_instance_id, goal_type, target_date_offset=DEFAULT_TARGET_DATE_OFFSET
    case goal_type
    when "G[T][T]"
        #Goal with on single-concept graph of concept T, with:
        #   - a single arbitrary assessment module within concept T
        #   - all modules under concept T recommendable.

        #select the single concept T
        concept = find_concept_by_name graph, 'T'

        #Pick the single assessment module under concept T.
        assessing_module = find_associated_assessing_bndl graph, concept['id']

        #Pick the all modules under concept T as recommendable.
        rec_modules = find_associated_assessing_atoms graph, concept['id']

        goal = create_goal learning_instance_id, {
            'target_modules' => [ {
                'module_id' => assessing_module,
                'target_score' => DEFAULT_TARGET_SCORE,
                'target_date' => days_from_now(target_date_offset),
            }],
            'recommendable_modules' => rec_modules.map{|m| { "module_id" => m } },
        }
    when "G[T][P1,T]"
        #Goal with on single-concept graph of concept T, with:
        #   - a single arbitrary assessment module within concept T
        #   - all modules under concepts P1 and T recommendable.

        #select concepts
        concept_t = find_concept_by_name graph, 'T'
        concept_p1 = find_concept_by_name graph, 'P1'

        #Pick the single assessment module under concept T.
        assessing_module = find_associated_assessing_bndl graph, concept_t['id']

        #Pick the all modules under concept T as recommendable.
        rec_modules = []
        rec_modules.concat( find_associated_assessing_atoms graph, concept_t['id'] )
        rec_modules.concat( find_associated_assessing_atoms graph, concept_p1['id'] )

        goal = create_goal learning_instance_id, {
            'target_modules' => [ {
                'module_id' => assessing_module,
                'target_score' => DEFAULT_TARGET_SCORE,
                'target_date' => days_from_now(target_date_offset),
            }],
            'recommendable_modules' => rec_modules.map{|m| { "module_id" => m } },
        }
    when "G[T][all]"
        #Goal with on single-concept graph of concept T, with:
        #   - a single arbitrary assessment module within concept T
        #   - all atoms in the graph recommendable.

        #select concepts
        concept_t = find_concept_by_name graph, 'T'

        #Pick the single assessment module under concept T.
        assessing_module = find_associated_assessing_bndl graph, concept_t['id']

        #Pick the all atoms as recommendable.
        rec_modules = find_all_atom_ids graph

        goal = create_goal learning_instance_id, {
            'target_modules' => [ {
                'module_id' => assessing_module,
                'target_score' => DEFAULT_TARGET_SCORE,
                'target_date' => days_from_now(target_date_offset),
            }],
            'recommendable_modules' => rec_modules.map{|m| { "module_id" => m } },
        }
    when "G[P1][all]"
        #Goal with on single-concept graph of concept P1, with:
        #   - a single arbitrary assessment module within concept P1
        #   - all atoms in the graph recommendable.

        #select concepts
        concept_p1 = find_concept_by_name graph, 'P1'

        #Pick the single assessment module under concept P1.
        assessing_module = find_associated_assessing_bndl graph, concept_p1['id']

        #Pick the all atoms as recommendable.
        rec_modules = find_all_atom_ids graph

        goal = create_goal learning_instance_id, {
            'target_modules' => [ {
                'module_id' => assessing_module,
                'target_score' => DEFAULT_TARGET_SCORE,
                'target_date' => days_from_now(target_date_offset),
            }],
            'recommendable_modules' => rec_modules.map{|m| { "module_id" => m } },
        }
    else
        throw "Unrecognized goal type #{goal_type}"
    end
end

def complete_goal_with_percentage_correct registration_id, goal_id, percentage_correct, options={}
    starting_event_threshold = 1.0 - percentage_correct

    completion_data = complete_goal registration_id,
        goal_id,
        options.merge({:starting_event_threshold=>starting_event_threshold, :event_threshold_multiplier=>1.0})

    completion_data
end

def complete_goal_with_percentage_correct_per_concept graph, registration_id, goal_id, percentages_correct, options={}
    concept_thresholds = []

    percentages_correct.each do |concept,percentage|
        concept_thresholds << {
            'id'=> (find_concept_by_name graph, concept)['id'],
            'current_threshold' => (1.0-percentage),
            'threshold_multiplier' => 1.0,
            'num_fails_before_success' => 0,
        }
    end

    completion_data = complete_goal registration_id,
        goal_id,
        options.merge({:concept_thresholds=>concept_thresholds})

    completion_data
end

#Parses test data as used in the recommendations contracts tests, stripping irrelevent data and applying readable names
def humanize_completion_data test_data, graph_type, scenario_title, scenario
    human_data = {}
    human_data['date'] = Time.now
    human_data['env'] = ENV['GRIFTER_ENVIRONMENT']
    human_data['contract'] = scenario_title
    human_data['graph_type'] = graph_type
    human_data['goal_type'] = scenario['goal']
    human_data['is_goal_complete'] = test_data['completion_data']['event_count'] <= scenario['max_adaptive_cycles']
    human_data['contract_vars'] = test_data['contract_vars']
    if test_data['preliminary_event_bodies']
        human_data['preliminary_events'] = test_data['preliminary_event_bodies'].map do |event|
            humanize_event_body test_data['graph'], event
        end
    end
    human_data['adaptive_cycles'] = test_data['completion_data']['event_bodies_sent'].map do |event|
        humanize_event_body test_data['graph'], event
    end
    human_data['concepts'] = humanize_concept_data test_data['graph']

    human_data
end

def humanize_event_body graph, event
    node = get_node_with_concept_and_content_type graph, event['module_id']
    cycle_data = {
        'concept' => node['concept']['name'],
        'content_type' => node['content_type'],
        'is_correct' => event['is_correct'],
        'module_id' => event['module_id'],
    }
end

def humanize_concept_data graph
    concepts = (find_all_concept_ids graph).map do |concept|
        concept = get_node_by_id graph, concept
        {'id' => concept['id'], 'name' => concept['name']}
    end
end

def write_contract_yaml_report contract_filename, report_data
    File.open("reports/contract_#{contract_filename}.yml", 'w') {|f| f.write YAML::dump(report_data) }
end
