# Generic methods for writing some reports
# The reports that can be made are:
# - all_requests_report  - a simple csv noting some details about every request
# - requests by endpoint - compiles response time stats about each endpoint that has been used
#

require 'fileutils'

# construct a path to the report directory, and ensure it exists
def report_dir
  report_dir = File.join(Dir.pwd, 'reports')
  FileUtils.mkdir_p report_dir
  report_dir
end

# write a csv that lists all requests
def write_all_requests_report filename=nil
  # default the filename to reports/all_requests.csv
  filename = File.join(report_dir, 'all_requests.csv') unless filename

  requests = metrics_all_requests
  Log.info "Writing all #{requests.length} requests report to filename: #{filename}"

  columns = [:service, :method, :path, :status, :duration_ms, :end_time]
  fmt_str = '%-8.8s,%-8.8s,%-61.60s,%-8.8s,%-12.12s,%-28.28s'
  File.open(filename, 'w') do |output|
    output.puts(fmt_str % columns)
    requests.each do |data|
      output.puts(fmt_str % data.to_a)
    end
  end
  Log.debug "Done writing all request report"
  filename
end


# write a csv that lists all unique endpoints
# and for each endpoint, some stats about response time
def write_metrics_by_endpoint_report filename=nil
  filename = File.join(report_dir, 'metrics_by_endpoint.csv') unless filename

  requests = metrics_all_requests

  # now compile the list of requests into a list of specific endoints that were hit
  # this means substituting specific IDs in requests paths with the string ':id'
  metrics_by_endpoint = {}
  requests.each do |sample|
    method = sample.method.upcase  # GET, POST, PUT, etc...
    path = sample.path.gsub(/\?.*$/,'') # the path, chop off query params

    #replace uuids with :id
    path.gsub!(/[0-9a-f\-]{36}/,':id')
    #replace any url path component with a capital letter, number or % to :id
    path.gsub!(/[^\/]+[A-Z0-9% ][^\/]+/,':id')

    endpoint = "#{'%-7s' % method} #{path}"

    metrics_by_endpoint[endpoint] ||= {
      durations: [],
      count: 0
    }
    metrics_by_endpoint[endpoint][:count] += 1
    metrics_by_endpoint[endpoint][:durations] << sample.duration_ms
  end

  Log.info "Writing metrics by endpoint report for #{metrics_by_endpoint.length} endpoints to filename: #{filename}"
  File.open(filename, 'w') do |output|
    columns = [:endpoint, :count, :mean_duration, :median_duration, :standard_deviation, :min_duration, :percentile_95, :max_duration]
    fmt_str = '%-80s' + columns[1..-1].inject(''){|s, c| s + ",%-#{c.length+1}s"}
    output.puts fmt_str % columns

    metrics_by_endpoint.each_pair do |endpoint, data|
      durations = data[:durations].sort
      output.puts fmt_str % [endpoint, data[:count], durations.mean.to_i, durations[durations.length/2], durations.sample_standard_deviation.to_i, durations.first, durations.percentile(0.95).to_i, durations.last]
    end
  end

  Log.debug "Done writing metrics by endpoint report"
  filename
end


# MONKEY PATCH ARRAY:
# methods for doing basic stats calculations on Arrays
# used in above metrics_by_endpoint_report method
class ::Array
  def sum
    self.inject(:+).to_f
  end

  def mean
    return 0.0 if self.length < 1
    self.sum / self.length.to_f
  end

  def sample_standard_deviation
    #unless we have at least 2 items in list, std dev is 0 by convention
    return 0.0 if self.length < 2

    mean = self.mean
    sum = self.inject(0.0) { |acc, i| acc + ((i - mean)**2) }

    Math.sqrt( sum / (self.length - 1))
  end

  def standard_error_of_the_mean
    return 0.0 if self.length < 2
    (self.sample_standard_deviation / Math.sqrt(self.length))
  end

  def percentile(percentile)
    return 0.0 if self.length < 1
    return self[0] if self.length < 2
    values_sorted = self.sort
    k = (percentile*(values_sorted.length-1)+1).floor - 1
    f = (percentile*(values_sorted.length-1)+1).modulo(1)
    return values_sorted[k] + (f * (values_sorted[k+1] - values_sorted[k]))
  end
end



