def set_account arg1, arg2=nil
  Log.debug "Requested user '#{arg1}' with credentials '#{arg2}'"
  #find the correct set of user and client credentials
  account={}
  case arg1
  when Symbol
    account = get_account arg1
    account.delete('client')
  when Hash
    if arg1['external_user_id'] and arg1['client_credentials']
      account = arg1
    end
  when String
    account['external_user_id'] = arg1
    case arg2
    when Symbol
      account['client_credentials'] = get_client_credentials arg2
    when Hash
      account['client_credentials'] = arg2
    end
  else raise TypeError.new('First argument must be a Symbol, Hash, or String')
  end
  unless account['external_user_id'] and account['client_credentials']
    raise ArgumentError.new('Could not set user and client credentials')
  end
  Log.debug "Resolved to account user '#{account['external_user_id']}' with partner '#{account['client_credentials']['client_id']}'"

  #get access token for account (and authorize if first time making account)
  token_request_params = {
    'code' => nil,
    'redirect_uri' => nil,
    'scope' => account['external_user_id'],
    'grant_type' => "client_credentials",
    'client_credentials' => account['client_credentials'],
    'external_user_id' => account['external_user_id']
  }
  access_token = do_cached_token_request token_request_params

  set_authorization_headers access_token
  true
end

# Cache access tokens to improve performance when switching between test users rapidly
# All the cacheing logic lives here
# We also check that the cache tokens are not expired and refresh them if they are
TOKEN_CACHE={}
def do_cached_token_request token_request_params, options={}

  # build a cache key and see if anything in the cache
  cache_key = token_request_params.select{|k| ['scope', 'grant_type', 'external_user_id', 'client_credentials'].include?(k) }
  tokens = TOKEN_CACHE[cache_key]

  #nothing was in the cache, so just do the token request
  if tokens.nil?
    Log.debug "Token cache miss: will do token request"
    tokens = get_access_token token_request_params
    tokens['resolved_token_expiration_time'] = Time.now + tokens['expires_in']
    TOKEN_CACHE[cache_key] = tokens
    seconds_remaining = tokens['resolved_token_expiration_time'] - Time.now
    Log.debug "Get token for this account:\ntoken: #{tokens['access_token']}\nexternal_user_id: #{cache_key['external_user_id']}\nseconds remaining: #{seconds_remaining.to_i}"

  # the cache had the needed tokens, but we need to check if they have expired
  else
    seconds_remaining = tokens['resolved_token_expiration_time'] - Time.now
    Log.debug "Token cache hit: found cached token response for this account:\ntoken: #{tokens['access_token']}\nexternal_user_id: #{cache_key['external_user_id']}\nseconds remaining: #{seconds_remaining.to_i}"

    # Bivins uses 60 seconds expiry for it's own server side cache of access tokens.
    # We might as well use the same cache timeout client side since otherwise the refresh will simply return the server cached tokens
    if seconds_remaining < 60
      Log.debug "cached token has less than 60 seconds until expiration, refreshing..."

      token_request_params.merge! 'refresh_token' => tokens['refresh_token'],
        'grant_type' => 'refresh_token'

      tokens = get_access_token_from_refresh_token token_request_params
      tokens['resolved_token_expiration_time'] = Time.now + tokens['expires_in']
      TOKEN_CACHE[cache_key] = tokens
    end
  end

  tokens['access_token']
end

# remove anything from the token cache for a given account id
# this is used when an account is deleted to clear the cache
def delete_account_id_from_token_cache account_id
  TOKEN_CACHE.delete_if {|k,v| v['account_id'] == account_id }
end

# temporarily authenticate as a user.
# Do the block.
# Go back to the orriginal user.
def as_account arg1, arg2=nil, &blk
  cur_auth_header = kapi.headers["Authorization"]
  Log.debug "as_account called, Authorization header currently set to '#{cur_auth_header}'"

  set_account arg1, arg2
  yield

ensure
  # whether the block returns, or threw exception, we need to change back to original user
  if cur_auth_header.nil?
    kapi.headers.delete "Authorization"
  else
    set_authorization_headers cur_auth_header
  end
  Log.debug "as_user over, Authorization header set back to '#{cur_auth_header}'"
end

def knerd_accounts
  #if were dealing with any kind of local environment, return some pre-defined constants
  if local_environment? || grifter_configuration[:environment].to_s.downcase =~ /local/

    knewton_client_credentials = {
      'client_id' => 'knewton',
      'client_secret' => '321knewton123',
    }

    {
      'knerd' => {
        'external_user_id' => 'knewton-system-user',
        'client' => 'knewton',
        'client_credentials' => knewton_client_credentials,
      },
      'prod_knerd' => {
        'external_user_id' => 'knewton-prod-system-user',
        'client' => 'knewton',
        'client_credentials' => knewton_client_credentials,
      },
    }

  else
    get_test_data :knerd_accounts
  end
end

def test_account_map
  @test_account_map ||= begin
                          full_data = get_test_data(:test_users_and_partners)
                          full_data['accounts'].merge! knerd_accounts
                          full_data['partners']['knewton'] = full_data['accounts']['knerd']['client_credentials']
                          full_data
                        end
  @test_account_map.clone  #clone to protect our original copy
end

def get_account label
  if label==:partner
    test_account_map['accounts']['system_user_default']
  elsif label==:user
    test_account_map['accounts']['user_default']
  else
    test_account_map['accounts'][label.to_s]
  end
end

def get_partner_helper label=nil
  if !label or label==:partner
    test_account_map['partners']['test_partner_default']
  else
    test_account_map['partners'][label.to_s]
  end
end

def get_client_id label=nil
  (get_partner_helper label)['client_id']
end

def get_client_secret label=nil
  (get_partner_helper label)['client_secret']
end

def get_client_credentials label=nil
  partner = get_partner_helper label
  {
    'client_id' => partner['client_id'],
    'client_secret' => partner['client_secret']
  }
end

def get_client_partner_id label=nil
  if label==nil or label==:partner
    test_account_map['partners']['test_partner_default']['partner_id']
  else
    test_account_map['partners'][label.to_s]['partner_id']
  end
end

#this method is used for grifters automatic authentication feature
#when calling grifts off the cmd line, this authentication will be
#done automatically first
def bivins_grifter_authenticate
  set_account :knerd
end

#this method is used to set a descriptive user-agent header before
#executing any method off the command line
def user_agent_grifter_authenticate
  user = Etc.getpwuid(Process.uid).name.slice(0,15)
  cmd_method_name = ARGV.first.slice(0,15)
  tracing_header = rand(2**32..2**64-1).to_s(16)
  Log.debug "Setting user-agent header to '#{cmd_method_name}', tracing header to #{tracing_header}"
  kapi.headers['user-agent'] = kapi.config[:default_headers][:"user-agent"] + " - #{cmd_method_name} - #{user}"
  kapi.headers['X-B3-TraceId'] = tracing_header
rescue Exception
  Log.warn "Failed to set method-specific user-agent header"
end

# create_partner_with_admin creates
# * a new partner with api key set to a name optionally
# * a partner admin for that partner
#
# returns a hash with all the data in a format that is ready to
# use with set_account / as_account methods.
def create_partner_with_admin options={}
  client_id = options.fetch('name', random_string(23)) + "-#{random_string(8)}"
  options.delete('name')
  client_secret = options.fetch('password', random_string)
  options.delete('password')

  as_account :knerd do
    #make the partner (and keep the partner id)
    partner_options = {'name' => client_id,
                       'password' => client_secret,
                       'rate_limit_multiplier' => 1.0,
                       'default_entitlements' => options['default_entitlements'],
    }.merge(options)
    partner = create_partner partner_options

    #make the system user
    system_user = create_system_user partner['id']

    {
      'account_id' => system_user['id'],
      'external_user_id' => system_user['external_user_id'],
      'partner_id' => partner['id'],
      'client_credentials' => {
        'client_id' => client_id,
        'client_secret' => client_secret,
      },
    }
  end
end

#create a user account, and get the account id
#This is useful for making fresh accounts that can help avoid entity limits
def create_test_account client_credentials
  external_user_id = create_account['external_user_id']
  account_id = get_account_id_for external_user_id, client_credentials
  {
    'account_id' => account_id,
    'external_user_id' => external_user_id,
  }
end

def get_partner_id_for_current_account
  orig_header_val = kapi.headers['X-Include-Partner-Id']
  kapi.headers['X-Include-Partner-Id'] = 'true'
  get_current_account
  pid = kapi.last_response['X-Partner-Id']
  kapi.headers['X-Include-Partner-Id'] = orig_header_val

  pid
end
